@@ -1,20 +1,6 @@
-/* Copyright (C) 2009-2022 Greenbone Networks GmbH
+/* SPDX-FileCopyrightText: 2009-2023 Greenbone AG
  *
  * SPDX-License-Identifier: GPL-2.0-or-later
- *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 /**
@@ -179,20 +165,32 @@ nvti_set_oid (nvti_t *, const gchar *);
 int
 nvti_set_name (nvti_t *, const gchar *);
 int
+nvti_put_name (nvti_t *, gchar *);
+int
 nvti_set_summary (nvti_t *, const gchar *);
 int
+nvti_put_summary (nvti_t *, gchar *);
+int
 nvti_set_insight (nvti_t *, const gchar *);
 int
+nvti_put_insight (nvti_t *, gchar *);
+int
 nvti_set_affected (nvti_t *, const gchar *);
 int
+nvti_put_affected (nvti_t *, gchar *);
+int
 nvti_set_impact (nvti_t *, const gchar *);
 int
+nvti_put_impact (nvti_t *, gchar *);
+int
 nvti_set_creation_time (nvti_t *, const time_t);
 int
 nvti_set_modification_time (nvti_t *, const time_t);
 int
 nvti_set_solution (nvti_t *, const gchar *);
 int
+nvti_put_solution (nvti_t *, gchar *);
+int
 nvti_set_solution_type (nvti_t *, const gchar *);
 int
 nvti_set_solution_method (nvti_t *, const gchar *);
@@ -217,6 +215,8 @@ nvti_set_required_udp_ports (nvti_t *, const gchar *);
 int
 nvti_set_detection (nvti_t *, const gchar *);
 int
+nvti_put_detection (nvti_t *, gchar *);
+int
 nvti_set_qod_type (nvti_t *, const gchar *);
 int
 nvti_set_qod (nvti_t *, const gchar *);
@@ -226,6 +226,8 @@ int
 nvti_set_category (nvti_t *, const gint);
 int
 nvti_set_family (nvti_t *, const gchar *);
+int
+nvti_put_family (nvti_t *, gchar *);
 
 int
 nvti_add_refs (nvti_t *, const gchar *, const gchar *, const gchar *);
