@@ -1,20 +1,6 @@
-/* Copyright (C) 2009-2022 Greenbone Networks GmbH
+/* SPDX-FileCopyrightText: 2009-2023 Greenbone AG
  *
  * SPDX-License-Identifier: GPL-2.0-or-later
- *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
 /**
@@ -743,6 +729,280 @@ try_read_entity_and_string (gnutls_session_t *session, int timeout,
     }
 }
 
+/**
+ * @brief Try read a response from a TLS session.
+ *
+ * @param[in]   session        Pointer to GNUTLS session.
+ * @param[in]   timeout        Server idle time before giving up, in seconds.  0
+ *                             to wait forever.
+ * @param[out]  string_return  An optional return location for the text read
+ *                             from the session.
+ *
+ * If string_return is NULL then it simply remains NULL.  If it is pointer to
+ * NULL then it points to a freshly allocated GString on successful return.
+ * Otherwise it must point to an existing GString onto which the text is
+ * appended.
+ *
+ * @return 0 success, -1 read error, -4 timeout.
+ */
+static int
+try_read_string (gnutls_session_t *session, int timeout,
+                 GString **string_return)
+{
+  GString *string;
+  int socket;
+  time_t last_time;
+  char *buffer; // Buffer for reading from the server.
+
+  /* Record the start time. */
+
+  if (time (&last_time) == -1)
+    {
+      g_warning ("   failed to get current time: %s\n", strerror (errno));
+      return -1;
+    }
+
+  if (timeout > 0)
+    {
+      /* Turn off blocking. */
+
+      socket = GPOINTER_TO_INT (gnutls_transport_get_ptr (*session));
+      if (fcntl (socket, F_SETFL, O_NONBLOCK) == -1)
+        return -1;
+    }
+  else
+    /* Quiet compiler. */
+    socket = 0;
+
+  buffer = g_malloc0 (BUFFER_SIZE);
+
+  /* Setup return arg. */
+
+  if (string_return == NULL)
+    string = NULL;
+  else if (*string_return == NULL)
+    string = g_string_new ("");
+  else
+    string = *string_return;
+
+  /* Read until encountering end of file or error. */
+
+  while (1)
+    {
+      ssize_t count;
+      int retries = 10;
+      while (1)
+        {
+          g_debug ("   asking for %i\n", BUFFER_SIZE);
+          count = gnutls_record_recv (*session, buffer, BUFFER_SIZE);
+          if (count < 0)
+            {
+              if (count == GNUTLS_E_INTERRUPTED)
+                /* Interrupted, try read again. */
+                continue;
+              if ((timeout > 0) && (count == GNUTLS_E_AGAIN))
+                {
+                  /* Server still busy, either timeout or try read again. */
+                  if ((timeout - (time (NULL) - last_time)) <= 0)
+                    {
+                      g_warning ("   timeout\n");
+                      if (fcntl (socket, F_SETFL, 0L) < 0)
+                        g_warning ("%s: failed to set socket flag: %s",
+                                   __func__, strerror (errno));
+                      g_free (buffer);
+                      return -4;
+                    }
+                  continue;
+                }
+              else if ((timeout == 0) && (count == GNUTLS_E_AGAIN))
+                {
+                  /* Server still busy, try read again.
+                   * If there is no timeout set and the server is still not
+                   * ready, it will try up to 10 times before closing the
+                   * socket. */
+                  if (retries > 0)
+                    {
+                      retries = retries - 1;
+                      continue;
+                    }
+                }
+
+              if (count == GNUTLS_E_REHANDSHAKE)
+                /* Try again. TODO Rehandshake. */
+                continue;
+              if (string && (*string_return == NULL))
+                g_string_free (string, TRUE);
+              if (timeout > 0)
+                {
+                  if (fcntl (socket, F_SETFL, 0L) < 0)
+                    g_warning ("%s: failed to set socket flag: %s", __func__,
+                               strerror (errno));
+                }
+              g_free (buffer);
+              return -1;
+            }
+          if (count == 0)
+            {
+              /* End of file. */
+              if (timeout > 0)
+                {
+                  if (fcntl (socket, F_SETFL, 0L) < 0)
+                    g_warning ("%s :failed to set socket flag: %s", __func__,
+                               strerror (errno));
+                }
+              if (string)
+                *string_return = string;
+              g_free (buffer);
+              return 0;
+            }
+          break;
+        }
+
+      g_debug ("<= %.*s\n", (int) count, buffer);
+
+      if (string)
+        g_string_append_len (string, buffer, count);
+
+      if ((timeout > 0) && (time (&last_time) == -1))
+        {
+          g_warning ("   failed to get current time (1): %s\n",
+                     strerror (errno));
+          if (fcntl (socket, F_SETFL, 0L) < 0)
+            g_warning ("%s :failed to set socket flag: %s", __func__,
+                       strerror (errno));
+          g_free (buffer);
+          return -1;
+        }
+    }
+}
+
+/**
+ * @brief Try read an XML entity tree from the socket.
+ *
+ * @param[in]   socket         Socket to read from.
+ * @param[in]   timeout        Server idle time before giving up, in seconds.  0
+ * to wait forever.
+ * @param[out]  entity         Pointer to an entity tree.
+ * @param[out]  string_return  An optional return location for the text read
+ *                             from the session.  If NULL then it simply
+ *                             remains NULL.  If a pointer to NULL then it
+ * points to a freshly allocated GString on successful return. Otherwise it
+ * points to an existing GString onto which the text is appended.
+ *
+ * @return 0 success, -1 read error, -2 parse error, -3 end of file, -4 timeout.
+ */
+static int
+try_read_string_s (int socket, int timeout, GString **string_return)
+{
+  GString *string;
+  time_t last_time;
+  /* Buffer for reading from the socket. */
+  char *buffer;
+
+  /* Record the start time. */
+
+  if (time (&last_time) == -1)
+    {
+      g_warning ("   failed to get current time: %s\n", strerror (errno));
+      return -1;
+    }
+
+  if (timeout > 0)
+    {
+      /* Turn off blocking. */
+
+      if (fcntl (socket, F_SETFL, O_NONBLOCK) == -1)
+        return -1;
+    }
+
+  buffer = g_malloc0 (BUFFER_SIZE);
+
+  /* Setup return arg. */
+
+  if (string_return == NULL)
+    string = NULL;
+  else if (*string_return == NULL)
+    string = g_string_sized_new (8192);
+  else
+    string = *string_return;
+
+  /* Read until encountering end of file or error. */
+
+  while (1)
+    {
+      int count;
+      while (1)
+        {
+          g_debug ("   asking for %i\n", BUFFER_SIZE);
+          count = read (socket, buffer, BUFFER_SIZE);
+          if (count < 0)
+            {
+              if (errno == EINTR)
+                /* Interrupted, try read again. */
+                continue;
+              if (timeout > 0)
+                {
+                  if (errno == EAGAIN)
+                    {
+                      /* Server still busy, either timeout or try read again. */
+                      if ((timeout - (time (NULL) - last_time)) <= 0)
+                        {
+                          g_warning ("   timeout\n");
+                          if (fcntl (socket, F_SETFL, 0L) < 0)
+                            g_warning ("%s :failed to set socket flag: %s",
+                                       __func__, strerror (errno));
+                          g_free (buffer);
+                          if (string && *string_return == NULL)
+                            g_string_free (string, TRUE);
+                          return -4;
+                        }
+                    }
+                  continue;
+                }
+              if (string && *string_return == NULL)
+                g_string_free (string, TRUE);
+              if (timeout > 0)
+                fcntl (socket, F_SETFL, 0L);
+              g_free (buffer);
+              return -1;
+            }
+          if (count == 0)
+            {
+              /* End of file. */
+              if (timeout > 0)
+                {
+                  if (fcntl (socket, F_SETFL, 0L) < 0)
+                    g_warning ("%s :failed to set socket flag: %s", __func__,
+                               strerror (errno));
+                }
+              if (string)
+                *string_return = string;
+              g_free (buffer);
+              return 0;
+            }
+          break;
+        }
+
+      g_debug ("<= %.*s\n", (int) count, buffer);
+
+      if (string)
+        g_string_append_len (string, buffer, count);
+
+      if ((timeout > 0) && (time (&last_time) == -1))
+        {
+          g_warning ("   failed to get current time (1): %s\n",
+                     strerror (errno));
+          if (fcntl (socket, F_SETFL, 0L) < 0)
+            g_warning ("%s :failed to set server socket flag: %s", __func__,
+                       strerror (errno));
+          g_free (buffer);
+          if (string && *string_return == NULL)
+            g_string_free (string, TRUE);
+          return -1;
+        }
+    }
+}
+
 /**
  * @brief Try read an XML entity tree from the socket.
  *
@@ -1082,6 +1342,42 @@ read_entity_and_text_c (gvm_connection_t *connection, entity_t *entity,
   return read_entity_and_string_c (connection, entity, NULL);
 }
 
+/**
+ * @brief Read text from the server.
+ *
+ * @param[in]  connection  Connection.
+ * @param[out] text        A pointer to a pointer, at which to store the
+ *                         address of a newly allocated string holding the
+ *                         text read from the session.
+ *
+ * @return 0 success, -1 read error, -2 argument error.
+ */
+int
+read_text_c (gvm_connection_t *connection, char **text)
+{
+  GString *string;
+  int ret;
+
+  if (text == NULL)
+    return -2;
+
+  string = NULL;
+
+  if (connection->tls)
+    ret = try_read_string (&connection->session, 0, &string);
+  else
+    ret = try_read_string_s (connection->socket, 0, &string);
+
+  if (ret)
+    {
+      if (string)
+        g_string_free (string, TRUE);
+      return ret;
+    }
+  *text = g_string_free (string, FALSE);
+  return 0;
+}
+
 /**
  * @brief Read entity and text. Free the entity immediately.
  *
@@ -1704,7 +2000,8 @@ parse_element (const gchar *string, element_t *element)
   if (xmlMemSetup (g_free, g_malloc, g_realloc, g_strdup))
     return -4;
 
-  doc = xmlReadMemory (string, strlen (string), "noname.xml", NULL, 0);
+  doc =
+    xmlReadMemory (string, strlen (string), "noname.xml", NULL, XML_PARSE_HUGE);
   if (doc == NULL)
     return -2;
 
@@ -1901,7 +2198,12 @@ element_t
 element_first_child (element_t element)
 {
   if (element)
-    return element->children;
+    {
+      element = element->children;
+      while (element && (element->type != XML_ELEMENT_NODE))
+        element = element->next;
+      return element;
+    }
   return NULL;
 }
 
@@ -1916,6 +2218,94 @@ element_t
 element_next (element_t element)
 {
   if (element)
-    return element->next;
+    {
+      element = element->next;
+      while (element && (element->type != XML_ELEMENT_NODE))
+        element = element->next;
+      return element;
+    }
   return NULL;
 }
+
+/**
+ * @brief Output the XML element as a string.
+ *
+ * The generated XML string will include namespace definitions from ancestor
+ *  elements.
+ *
+ * @param[in]  element  The element to output as a string.
+ *
+ * @return The newly allocated XML string.
+ */
+gchar *
+element_to_string (element_t element)
+{
+  xmlBufferPtr buffer;
+  char *xml_string;
+
+  // Copy element to ensure XML namespaces are included
+  element_t element_copy;
+  element_copy = xmlCopyNode (element, 1);
+
+  buffer = xmlBufferCreate ();
+  xmlNodeDump (buffer, element_copy->doc, element_copy, 0, 0);
+  xmlFreeNode (element_copy);
+
+  xml_string = g_strdup ((char *) xmlBufferContent (buffer));
+
+  xmlBufferFree (buffer);
+  return xml_string;
+}
+
+/**
+ * @brief Print an XML element tree to a GString, appending it if string is not
+ * @brief empty.
+ *
+ * @param[in]      element  Element tree to print to string.
+ * @param[in,out]  string  String to write to.
+ */
+void
+print_element_to_string (element_t element, GString *string)
+{
+  gchar *text_escaped, *text;
+  element_t ch;
+  xmlAttr *attribute;
+
+  text_escaped = NULL;
+
+  g_string_append_printf (string, "<%s", element_name (element));
+
+  attribute = element->properties;
+  while (attribute)
+    {
+      xmlChar *value;
+
+      value = xmlNodeListGetString (element->doc, attribute->children, 1);
+
+      text_escaped = g_markup_escape_text ((gchar *) value, -1);
+      g_string_append_printf (string, " %s=\"%s\"", attribute->name,
+                              text_escaped);
+      g_free (text_escaped);
+
+      xmlFree (value);
+
+      attribute = attribute->next;
+    }
+
+  g_string_append_printf (string, ">");
+
+  text = element_text (element);
+  text_escaped = g_markup_escape_text (text, -1);
+  g_free (text);
+  g_string_append_printf (string, "%s", text_escaped);
+  g_free (text_escaped);
+
+  ch = element_first_child (element);
+  while (ch)
+    {
+      print_element_to_string (ch, string);
+      ch = element_next (ch);
+    }
+
+  g_string_append_printf (string, "</%s>", element_name (element));
+}
